type id = int

let id_gen () =
  let i = ref (-1) in
  fun () -> incr i; !i

module type M = sig
  type t
  type init
  type action_functional
  type action_ui
  val create : init -> t * action_ui
  val on_action_func : t -> action_functional -> t * action_ui
end

module type FRP = sig
  type 'a t
  val create : unit -> 'a t * ('a -> unit)
  val fold : ('a -> 'b -> 'a) -> 'a -> 'b t -> 'a t
  val map : ('a -> 'b) -> 'a t -> 'b t
end

class ['a] callbacks ?(callbacks=[]) () = object
  val mutable cbs : ('a -> unit) list = callbacks
  method set cbs_new = cbs <- cbs_new
  method get = cbs
  method add cb = cbs <- cb :: cbs
  method remove cb = cbs <- List.filter ((!=) cb) cbs
  method call arg = List.iter (fun f -> f arg) cbs
end

module Core (FRP : FRP) (M : M) = struct
  class core ?state_callbacks ?msg_callbacks x =
    let e, e_send = FRP.create () in
    let state_init = M.create x in
    let msg_cb = new callbacks ?callbacks:msg_callbacks () in
    let state_cb = new callbacks ?callbacks:state_callbacks () in
    let state = FRP.fold (fun (t, _) e -> M.on_action_func t e) state_init e in
    let ref_p = FRP.map (fun ((t, msg) as x) -> state_cb#call x) state in
    let ref_cb = FRP.map (fun (_, msg) -> msg_cb#call msg) state in
    object(self)
      val references = fun () -> ignore (state, ref_cb, ref_p)
      method send = e_send
      method state_callbacks = state_cb
      method msg_callbacks = msg_cb
      initializer
        state_cb#call state_init
    end
end
